from typing import Optional

from marshmallow.fields import Field

from microcosm_flask.swagger.parameters.base import ParameterBuilder


SWAGGER_TYPE = "__swagger_type__"
SWAGGER_FORMAT = "__swagger_format__"


class DecoratedParameterBuilder(ParameterBuilder):
    """
    Build parameter from custom attributes (injected by a decorator).

    """
    def supports_field(self, field: Field) -> bool:
        return hasattr(field, SWAGGER_TYPE)

    def parse_format(self, field: Field) -> Optional[str]:
        return getattr(field, SWAGGER_FORMAT, None)

    def parse_type(self, field: Field) -> Optional[str]:
        field_type = getattr(field, SWAGGER_TYPE)
        if isinstance(field_type, list):
            # Ideally we'd use oneOf here, but OpenAPI 2.0 uses the 0.4-draft jsonschema
            # which doesn't include oneOf.
            return None
        return field_type



import sys
from datetime import datetime
from os import makedirs
from os.path import exists, expanduser, join

from flask import g


try:
    from pyinstrument import Profiler
except ImportError:
    pass


def default_profile_dir(name):
    return expanduser("~/.{name}/profile".format(name=name))


def enable_profiling(graph):
    if "pyinstrument.profiler" not in sys.modules:
        raise Exception("Profiling requires 'pyinstrument'. Make sure it is installed.")

    profile_dir = graph.config.flask.profile_dir or default_profile_dir(name=graph.metadata.name)
    if not exists(profile_dir):
        makedirs(profile_dir)

    graph.app.config['PROFILE'] = True

    app = graph.flask

    @app.before_request
    def before_request():
        g.profiler = Profiler()
        g.profiler.start()

    @app.after_request
    def after_request(response):
        g.profiler.stop()
        output_html = g.profiler.output_html()

        time_tag = datetime.now().strftime("%y-%m-%d-%H-%M-%S")
        profile_path = join(profile_dir, f"profile-{time_tag}.html")
        with open(profile_path, "w") as f:
            f.write(output_html)
        graph.app.logger.info(f"Profile saved into: {profile_path}")

        return response

    graph.app.logger.info(f"*** Profiling is ON, Will save profiling data to directory: {profile_dir}")



try:
    from microcosm_metrics.classifier import Classifier
except ImportError:
    raise Exception("Route metrics require 'microcosm-metrics'")

from microcosm_flask.audit import parse_response
from microcosm_flask.errors import extract_status_code


class StatusCodeClassifier(Classifier):
    """
    Label route result/error with its status code.

    """
    def label_result(self, result):
        _, status_code, _ = parse_response(result)
        return str(status_code)

    def label_error(self, error):
        status_code = extract_status_code(error)
        return str(status_code)



from flask import jsonify

from microcosm_flask.formatting.base import BaseFormatter


class JSONFormatter(BaseFormatter):

    CONTENT_TYPE = "application/json"

    @property
    def content_type(self):
        return JSONFormatter.CONTENT_TYPE

    def build_response(self, response_data):
        return jsonify(response_data)



from microcosm_flask.formatting.base import BaseFormatter


class TextFormatter(BaseFormatter):

    CONTENT_TYPE = "text/plain"

    @property
    def content_type(self):
        return TextFormatter.CONTENT_TYPE

    def format(self, response_data):
        return response_data



"""
A naming convention and discovery mechanism for HTTP endpoints.

Operations provide a naming convention for references between endpoints,
allowing easy construction of links or audit trails for external consumption.

"""
from collections import namedtuple
from enum import Enum, unique


# metadata for an operation
OperationInfo = namedtuple("OperationInfo", ["name", "method", "pattern", "default_code"])


# NB: Namespace.parse_endpoint requires that operation is the second argument
NODE_PATTERN = "{subject}.{operation}.{version}"
EDGE_PATTERN = "{subject}.{operation}.{object_}.{version}"


@unique
class Operation(Enum):
    """
    An enumerated set of operation types, which know how to resolve themselves into
    URLs and hrefs.

    """
    # discovery operation
    Discover = OperationInfo("discover", "GET", NODE_PATTERN, 200)

    # collection operations
    Search = OperationInfo("search", "GET", NODE_PATTERN, 200)
    Count = OperationInfo("count", "HEAD", NODE_PATTERN, 200)
    Create = OperationInfo("create", "POST", NODE_PATTERN, 201)
    DeleteBatch = OperationInfo("delete_batch", "DELETE", NODE_PATTERN, 204)
    UpdateBatch = OperationInfo("update_batch", "PATCH", NODE_PATTERN, 200)
    CreateCollection = OperationInfo("create_collection", "POST", NODE_PATTERN, 200)
    SavedSearch = OperationInfo("saved_search", "POST", NODE_PATTERN, 200)

    # instance operations
    Retrieve = OperationInfo("retrieve", "GET", NODE_PATTERN, 200)
    Delete = OperationInfo("delete", "DELETE", NODE_PATTERN, 204)
    Replace = OperationInfo("replace", "PUT", NODE_PATTERN, 200)
    Update = OperationInfo("update", "PATCH", NODE_PATTERN, 200)
    Alias = OperationInfo("alias", "GET", NODE_PATTERN, 302)

    # relation operations
    CreateFor = OperationInfo("create_for", "POST", EDGE_PATTERN, 201)
    DeleteFor = OperationInfo("delete_for", "DELETE", EDGE_PATTERN, 204)
    ReplaceFor = OperationInfo("replace_for", "PUT", EDGE_PATTERN, 200)
    RetrieveFor = OperationInfo("retrieve_for", "GET", EDGE_PATTERN, 200)
    SearchFor = OperationInfo("search_for", "GET", EDGE_PATTERN, 200)
    UpdateFor = OperationInfo("update_for", "PATCH", EDGE_PATTERN, 200)

    # file upload operations
    Upload = OperationInfo("upload", "POST", NODE_PATTERN, 200)
    UploadFor = OperationInfo("upload_for", "POST", EDGE_PATTERN, 200)

    # ad hoc operations
    Command = OperationInfo("command", "POST", NODE_PATTERN, 200)
    Query = OperationInfo("query", "GET", NODE_PATTERN, 200)

    @classmethod
    def from_name(cls, name):
        for operation in cls:
            if operation.value.name.lower() == name.lower():
                return operation
        else:
            raise ValueError(name)

    @property
    def endpoint_pattern(self):
        """
        Convert the operation's pattern into a regex matcher.

        """
        parts = self.value.pattern.split(".")
        return "[.]".join(
            "(?P<{}>[^.]*)".format(part[1:-1])
            for part in parts
        )



"""
Support for registering function metadata.

"""
from werkzeug.exceptions import InternalServerError
from werkzeug.routing import parse_rule

from microcosm_flask.namespaces import Namespace


REQUEST = "__request__"
RESPONSE = "__response__"
QS = "__qs__"


def iter_endpoints(graph, match_func):
    """
    Iterate through matching endpoints.

    The `match_func` is expected to have a signature of:

        def matches(operation, ns, rule):
            return True

    :returns: a generator over (`Operation`, `Namespace`, rule, func) tuples.

    """
    for rule in graph.flask.url_map.iter_rules():
        try:
            operation, ns = Namespace.parse_endpoint(rule.endpoint, get_converter(rule))
        except (IndexError, ValueError, InternalServerError):
            # operation follows a different convention (e.g. "static")
            continue
        else:
            # match_func gets access to rule to support path version filtering
            if match_func(operation, ns, rule):
                func = graph.flask.view_functions[rule.endpoint]
                yield operation, ns, rule, func


def get_converter(rule):
    """
    Parse rule will extract the converter from the rule as a generator

    We iterate through the parse_rule results to find the converter
    parse_url returns the static rule part in the first iteration
    parse_url returns the dynamic rule part in the second iteration if its dynamic

    """
    for converter, _, _ in parse_rule(str(rule)):
        if converter is not None:
            return converter
    return None


def request(schema):
    """
    Decorate a function with a request schema.

    """
    def wrapper(func):
        setattr(func, REQUEST, schema)
        return func
    return wrapper


def response(schema):
    """
    Decorate a function with a response schema.

    """
    def wrapper(func):
        setattr(func, RESPONSE, schema)
        return func
    return wrapper


def qs(schema):
    """
    Decorate a function with a query string schema.

    """
    def wrapper(func):
        setattr(func, QS, schema)
        return func
    return wrapper


def get_request_schema(func):
    return getattr(func, REQUEST, None)


def get_response_schema(func):
    return getattr(func, RESPONSE, None)


def get_qs_schema(func):
    return getattr(func, QS, None)



"""
Hamcrest matching support for JSON responses.

"""
from json import dumps, loads

from hamcrest.core.base_matcher import BaseMatcher

from microcosm_flask.fields.uri_field import normalize_uri


def prettify(value):
    return dumps(
        value,
        sort_keys=True,
        indent=4,
        separators=(',', ': '),
    )


class JSON:
    """
    Dictionary wrapper with JSON pretty-printing for Hamcrest's description.

    """
    def __init__(self, dct):
        self.dct = dct

    def __getitem__(self, key):
        return self.dct[key]

    def get(self, key):
        return self.dct.get(key)

    def describe_to(self, description):
        description.append(prettify(self.dct))


def json_for(value):
    if not isinstance(value, (dict, list)):
        value = loads(value)
    return JSON(value)


class JSONMatcher(BaseMatcher):
    """
    Hamcrest matcher of a JSON encoded resource.

    Subclasses must define `_matcher` and invoke `assert_that` within a request
    context to ensure that Flask's `url_for` can be resolved.

    Example:

        with graph.app.test_request_context():
           assert_that(json(response.data), matches_myresource(expected))

    """
    def __init__(self, resource):
        self.resource = resource
        self.schema = self.schema_class()
        self.expected = self.schema.dump(self.resource)

    @property
    def schema_class(self):
        raise NotImplementedError

    def describe_to(self, description):
        description.append_text("expected {}".format(prettify(self.expected)))


class URIMatcher(BaseMatcher):
    """
    Hamcrest matcher for comparing URIs by canonicalizing them first.

    Example:

        with graph.app.test_request_context():
           assert_that(uri, matches_uri("https://canonical.url.com/path?a=1&b=2"))

    """
    def __init__(self, uri):
        self.uri = uri

    def _matches(self, item):
        if not isinstance(item, str):
            return False

        return normalize_uri(item) == normalize_uri(self.uri)

    def describe_to(self, description):
        description.append_text("expected URI: {}".format(self.uri))


def matches_uri(uri):
    return URIMatcher(uri)
